/*
  statusindicatorwidget.cpp

  Copyright (c) 2022 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#include "statusindicatorwidget.h"
#include <QPainter>

StatusIndicatorWidget::StatusIndicatorWidget(QWidget *parent, QSacnUniverse *universe)
    : QWidget{parent}
    , mUniverse(universe)
    , mBlinkState(true)
{
    setMinimumWidth(16);
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::MinimumExpanding);

    // initial state
    setToolTip(statusText());
    statusChanged();

    // status feedback
    connect(mUniverse, &QSacnUniverse::statusChanged,
            this, [this](){setToolTip(statusText());});
    connect(mUniverse, &QSacnUniverse::statusChanged,
            this, &StatusIndicatorWidget::statusChanged);
    connect(mUniverse, &QSacnUniverse::dataChanged,
            this, &StatusIndicatorWidget::dataActivity);
}


/**
 * @brief StatusIndicatorWidget::statusChanged
 */
void StatusIndicatorWidget::statusChanged()
{
    switch (mUniverse->status()) {
    case sACN::Universe::DMX_LOST:
    {
        mBlinkState = !mBlinkState;
        QTimer::singleShot(250, this, [this]() {
            statusChanged();
        });
        break;
    }
    default:
        mBlinkState = true;;
    }
    update();
}


/**
 * @brief StatusIndicatorWidget::dataActivity
 */
void StatusIndicatorWidget::dataActivity()
{
    if (!mBlinkState)
      return; // already in a blink state

    QTimer::singleShot(100, this, [this]() {
        mBlinkState = true;
        update();
    });
    mBlinkState = false;
    update();
}


/**
 * @brief StatusIndicatorWidget::statusText
 * @return
 */
const QString StatusIndicatorWidget::statusText() const
{
    switch (mUniverse->status())
    {
    case sACN::Universe::DMX_NULL:
        return tr("Pending");
    case sACN::Universe::DMX_ACTIVE:
        return tr("Active");
    case sACN::Universe::DMX_LOST:
        return tr("Lost");
    case sACN::Universe::sACN_TERMINATED:
        return tr("Terminated");
    default:
        return tr("Undefined");
    }
}


void StatusIndicatorWidget::paintEvent(QPaintEvent *)
{
    QColor stroke(127,127,127);     // neutral
    QColor fill(64,64,64);          // grey

    switch (mUniverse->status()) {
    case sACN::Universe::DMX_NULL:
        fill.setRgb(250,250,250);   // white
        break;
    case sACN::Universe::DMX_ACTIVE:
        fill.setRgb(32,255,64);     // green
        break;
    case sACN::Universe::DMX_LOST:
        fill.setRgb(255,255,64);    // amber
        break;
    case sACN::Universe::sACN_TERMINATED:
        fill.setRgb(255,32,32);     // red
        break;
    default:
        break;
    }
    fill.setAlpha(mBlinkState ? 255 : 42);

    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing);
    painter.setPen(stroke);
    painter.setBrush(fill);

    const int size = qMin(width(), height());
    switch (mUniverse->status()) {
    case sACN::Universe::DMX_NULL:
    case sACN::Universe::DMX_ACTIVE:
    case sACN::Universe::DMX_LOST:
    case sACN::Universe::sACN_TERMINATED:
    default:
        painter.drawEllipse((width()-size)/2, (height()-size)/2, size, size);   // circle
    }
}
