/*
  arbitratinguniverse.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "data.h"
#include <memory>
#include <unordered_map>
#include <shared_mutex>

namespace DMX {
  class Universe;
}

namespace sACN {

class Universe;

/**
 * @brief Priority based selection of multiple source universes
 *
 *
 * > \cite sACN 6.2.3.1 Multiple Sources at Highest Priority
 * >
 * > It is possible for there to be multiple sources, all transmitting data
 * > at the highest currently active priority for a given universe. When this
 * > occurs, receivers must handle these sources in some way.
 *
 * If multiple universes with highest priority are recieved, the universe
 * recieved most recently will be given dominance.
 *
 * > \cite sACN 6.2.3.2 Note on Merge and Arbitration Algorithms
 * >
 * > A process which selects between candidate sources based on some
 * > additional selection criterion is called arbitration.
 *
 * The universe with the highest priority will be determined to be the
 * dominant source.
 */
class UniverseArbitrator
{
public:
    explicit UniverseArbitrator(sACN::Universe*);

    void refresh();


    // DMX::Universe API:
    long age() const;
    uint8_t status() const;
    uint8_t slot(const uint16_t) const;
    double rxRate();

    // sACN::Universe API
    std::shared_ptr<DATA::data_header> metadata() const;
    bool isEditable() const;
    uint16_t activeSlots() const;
    const std::vector<DATA::data_header> sources() const;

    // Source universes
    bool hasSourceUniverse(const DATA::data_header&) const;
    std::shared_ptr<Universe> sourceUniverse(const DATA::data_header&);
    std::shared_ptr<Universe> addNewSource(const DATA::data_header&);
    void deleteSourceUniverse(const DATA::data_header&);
    std::shared_ptr<void> onSourceListChange(const std::function<void(DMX::Universe*)>);

    // hold-last-look
    void setHoldLastLook(const bool);
    bool getHoldLastLook() const;

    // merge-mode
    /**
     * @brief The MergeMode enum
     */
    enum MergeMode {
      MERGE_OTHER,
      MERGE_HTP,
      MERGE_LTP
    };
    void setMergeMode(const MergeMode);
    MergeMode getMergeMode() const;

protected:
  inline void doListChangeCallbacks();

private:
  void find_dominant_();
  void purge_stale_sources_();

  sACN::Universe * m_universe;
  std::unordered_map<DATA::data_header, std::shared_ptr<Universe>> sources_;
  std::vector<std::shared_ptr<void>> cb_tokens_; //!< dominant universe callback tokens
  std::vector<std::shared_ptr<void>> source_tokens_; //!< source universe status changes
  std::vector<std::weak_ptr<const std::function<void(DMX::Universe*)>>> cb_sourceListChange;  //!< list of calback functions
  bool hold_last_look_;
  MergeMode merge_mode_;
  std::weak_ptr<Universe> m_dominant;
  mutable std::shared_mutex mtx_arbitrate;
};

} // SACN namespace
